# -*- coding: UTF-8 -*-

###########################################################################
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# twistd -noy controle_vnc_serveur.py
#
# Serveur Perspective Broker pour les clients Scribe/Windows
#
###########################################################################

# modules habituels
import os, sys, shutil
sys.path.append('/usr/share/eole/controlevnc')

# Perspective Broker
from twisted.spread import pb

# xml_rpc
from twisted.web import xmlrpc

# serveur HTTP
from twisted.web import server, static

# le service
from twisted.application import service, internet
from twisted.python import log
log.FileLogObserver.timeFormat = "%Y/%m/%d %H:%M:%S %Z"

# gestion de l'annuaire
import ldap_utils
from scribe.ldapconf import SUPPORT_ETAB
# gestion des clients
from cliscribe import Cliscribe
from connexions import Connexions
import gest_sessions

# gestion des stations
import manage_stations

# gestion des blocage réseau
from blocage import Blocage

# gestion des devoirs
from config import activer_eop
import devoirs

# autorisations d'accès (basé sur IP ...)
import autorisations

from scribe.eoleldap import Ldap
from scribe.linker import _user_factory
from fichier.passwd import check_nb_min_classes

import xmlrpclib

# chargement de la clé secrète pour les appels xmlrpc
try:
    f=open("/etc/controlevnc/.secret","r")
    secret_key=f.readline().rstrip('\n\r')
    f.close()
except:
    log.msg('Impossible de lire le fichier /etc/controlevnc/.secret nécessaire pour EOP, la gestion des devoirs')
    secret_key=""


######################
#  appels distants   #
######################
class Echoer(pb.Root):
    """Classe contenant les fonctions distantes "remote_<func>" accessibles depuis les postes clients.
    Protection de certaines en fonction du groupe dont fait parti le login qui a ouvert une session Samba
    sur la station tentant d'accéder à la remote_fonction.
    """
    def __init__(self):
        self.blocage = Blocage()
        return

#    def rootObject(self, broker):
#        """récupération d'informations sur la connection (ex: IP)
#        """
#        self.p_broker = broker
#        return pb.Root.rootObject(self, broker)

    def remoteMessageReceived(self, broker, message, args, kw):
        #self.ip = self.p_broker.transport.getPeer().host
        self.p_broker = broker
        self.ip = broker.transport.getPeer().host
        return pb.Root.remoteMessageReceived(self, broker, message, args, kw)

    #######################
    # Fonctions distantes #
    #######################

    def remote_bonjour(self, ip, ret=None):
        """Appel de la fonction bonjour sur client B par client A
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_bonjour par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_bonjour par %s sur %s'%(self.ip, ip))
        d = Cliscribe(ip).appel()
        return d

    # Listes, des groupes, des connectés et des membres de groupes #
    def remote_connectes(self, groupes=None):
        """gestion controle-vnc
        """
        log.msg('Appel de la fonction remote_connectes (groupes=%s) par %s'%(groupes, self.ip))
        if SUPPORT_ETAB:
            user = Connexions().get_active_user(self.ip)
        else:
            user = None
        if groupes: return Connexions().get_connectes_groupes(groupes, user)
        else: return Connexions().sessions

    def remote_groupes(self):
        """renvoie la liste des groupes
        """
        log.msg('Appel de la fonction remote_groupes par %s'%self.ip)
        if SUPPORT_ETAB:
            user = Connexions().get_active_user(self.ip)
        else:
            user = None
        return ldap_utils.get_groupes(user=user)

    def remote_classes(self):
        """renvoie la liste des classes
        """
        log.msg('Appel de la fonction remote_classes par %s'%self.ip)
        return ldap_utils.get_classes()

    def remote_classes_et_groupes(self):
        """renvoie la liste des classes et des groupes de type "Groupe"
        """
        log.msg('Appel de la fonction remote_classes_et_groupes par %s'%self.ip)
        if SUPPORT_ETAB:
            user = Connexions().get_active_user(self.ip)
        else:
            user = None
        return ldap_utils.get_classes_et_groupes(user=user)

    def remote_membres(self, groupe, rgroupe=None):
        """renvoie la liste des membres de 'groupe'
        """
        log.msg('Appel de la fonction remote_membres par %s (%s, %s)'%(self.ip, groupe, rgroupe))
        if rgroupe: return ldap_utils.get_membres_croise([groupe, rgroupe])
        else: return ldap_utils.get_membres(groupe)

    # Observation VNC #
    def remote_get_vnc_mode(self):
        log.msg('Appel de la fonction remote_get_vnc_mode par %s'%(self.ip))
        return gest_sessions.Logon().get_vnc_mode()

    def remote_observe(self, cible, active=None):
        """src est en mode "listen"
        clibe "connect" src, sur cible : 'winvnc.exe -connect src (self.ip)'
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_observe par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_observe par %s sur %s (active = %s)'%(self.ip, cible, active))
        # force le mode d'observation fixé dans l'EAD
        if not autorisations.isadmin(self.ip) and gest_sessions.Logon().get_vnc_mode() == 'simple': active=None
        d = Cliscribe(cible).vnc(action='setinputs', value=active, restart=True)
        d.addCallback(Cliscribe(cible).vnc, action='connect', value=self.ip)
        return d

    # Diffusion VNC #
    def remote_diffusion(self, ip_liste):
        """active le mode vncviewer /listen sur les ips de ip_liste
        """
        if not ip_liste: return
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_diffusion par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_diffusion par %s sur %s'%(self.ip, ip_liste))
        # fichier généré par vncviewer.exe configuration en 256 couleurs
        myconf = file('/home/workgroups/professeurs/gestion-postes/eole_conf.vnc').read()
        # winvnc en mode lecture seule (fait dans la session)
        #d = Cliscribe(self.ip).vnc(action='setinputs', value=None)
        d = Cliscribe(ip_liste[0]).vnc(action='start_listen', value='/viewonly', conf=myconf)
        for cible in ip_liste[1:]:
            d.addCallback(Cliscribe(cible).vnc, action='start_listen', value='/viewonly', conf=myconf)
            d.addErrback(Cliscribe(cible).vnc, action='start_listen', value='/viewonly', conf=myconf)
        return d

    # Blocage#
    def remote_get_bloc_list(self):
        """Renvoie la liste des type de blocages disponibles (Aucun=0, Tout internet=1, ...)
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_get_bloc_list par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_get_bloc_list par %s'%self.ip)
        return self.blocage.get_bloc_list()

    # ... par Machine
    def remote_get_machines_blocmod(self):
        """renvoie la liste des machines avec leur blocage actuel
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_get_machines_blocmod par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_get_machines_blocmod par %s'%self.ip)
        return self.blocage.get_computers_mod()

    def remote_set_machine_bloc_mod(self, machines):
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_set_machine_bloc_mod par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_set_machine_bloc_mod par %s, Machines=%s'%(self.ip, machines))
        return self.blocage.bloc_computers(machines)

    # ... par Utilisateur
    def remote_users_bloc_mod(self, groupe):
        """renvoie un dictionnaire des membres du groupe avec leur mode de blocage
        { user1 : {'netmod' : 1, 'partmod' : 0}, user2 :  ... }
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_users_bloc_mod par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_users_bloc_mod par %s GRP=%s'%(self.ip, groupe))
        if SUPPORT_ETAB:
            user = Connexions().get_active_user(self.ip)
        else:
            user = None
        return self.blocage.get_users_mod(groupe, user=user)

    def remote_set_user_bloc_mod(self, users):
        """Applique le blocage demandé aux utilisateurs sélectionnés
        users = [(user, netmod, partmod, durée du blocage), (user2, ...)]
        """
        if not autorisations.isprof(self.ip):
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_set_users_bloc_mod par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_set_user_bloc_mod par %s, Users=%s'%(self.ip, users))
        return self.blocage.bloc_users(users)

    # Distribution de devoirs
    def remote_nb_membres(self, groupe):
        """Renvoie le nombre de membres du groupe
        """
        log.msg('Appel de la fonction remote_nb_membres par %s pour %s'%(self.ip, groupe))
        if SUPPORT_ETAB:
            user = Connexions().get_active_user(self.ip)
        else:
            user = None
        return ldap_utils.get_nb_membres(groupe, user=user)

    def remote_get_devoirs(self, user=None):
        username = autorisations.isprof(self.ip)
        #FIXME
        if username == True:
            username = user
        if not username:
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_get_devoirs par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_get_devoirs par %s : username=%s'%(self.ip, username))
        return devoirs.get_devs(username)

    def remote_distribute(self, groupe, dev_name, eleve_only, in_perso, user=None):
        """distribue le devoir "dev_name" du prof "username" à "groupe"
        """
        username = autorisations.isprof(self.ip)
        #FIXME
        if username == True:
            username = user
        if not username:
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_distribute par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_distribute par %s : username=%s, groupe=%s, dev_name=%s, eleve_only=%s, %s'%(self.ip, username, groupe, dev_name, eleve_only, in_perso))
        return devoirs.distribute(username, groupe, dev_name, eleve_only, in_perso)

    def remote_ramasse(self, dev_name, user=None):
        """rammasse "dev_name"
        """
        username = autorisations.isprof(self.ip)
        #FIXME
        if username == True:
            username = user
        if not username:
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_ramasse par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_ramasse par %s : username=%s, dev_name=%s'%(self.ip, username, dev_name))
        return devoirs.ramasse(username, dev_name)

    def remote_rendre(self, dev_name, user=None):
        """rend la correction de "dev_name"
        """
        username = autorisations.isprof(self.ip)
        #FIXME
        if username == True:
            username = user
        if not username:
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_rendre par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_rendre par %s : username=%s, dev_name=%s'%(self.ip, username, dev_name))
        return devoirs.rendre(username, dev_name)

    def remote_supprimer(self, dev_name, user=None):
        """Supprime dev_name de la liste des devoirs et supprime les données correspondantes
        (sinon l'élève ne peut supprimer le dossier du devoir dans son rep perso
        """
        username = autorisations.isprof(self.ip)
        #FIXME
        if username == True:
            username = user
        if not username:
            self.p_broker.transport.loseConnection()
            log.msg('Appel non autorisé (non prof) de remote_supprimer par %s'%self.ip)
            return
        log.msg('Appel de la fonction remote_supprimer par %s : username=%s, dev_name=%s'%(self.ip, username, dev_name))
        return devoirs.supprimer(username, dev_name)

    # Gestion des postes (démarrage) #
    def remote_service_start(self, nom, ost=None, mac=None):
        """gestion des postes allumés
        """
        log.msg('Appel de la fonction remote_service_start par %s : %s (%s), mac=%s'%(self.ip, nom, ost, mac))
        return manage_stations.add_sta(self.ip, nom, ost, mac)

    # Gestion des postes (arrêt) #
    def remote_service_stop(self):
        log.msg('Appel de la fonction remote_service_stop par %s'%self.ip)
        return manage_stations.del_sta(self.ip)

    # Gestion de sessions cliente #
    # Ouverture de session
    def remote_logon(self, os_type):
        """fonction appelée par les clients lors de l'ouverture de session
        (dés)active le blocage
        """
        log.msg('Appel de la fonction remote_logon par %s'%(self.ip))
        return gest_sessions.Logon().logon(self.ip, os_type, self.blocage)
#        return True


class RpcServer(xmlrpc.XMLRPC):
    def verif_cle(self, cle):
        """Vérifie une clé en la comparant à celle dans config
        """
        return cle == secret_key

    def est_prof(self, username):
        """Vérifie si username est un prof
        """
        return ldap_utils.is_prof(username)

    def xmlrpc_remote_est_prof(self, cle, username):
        """renvoie True si le username est prof et Flase sinon
        """
        return self.est_prof(username)

    def xmlrpc_remote_groupes(self, groupname):
        """renvoie la liste des groupes
        """
        log.msg('Appel de la fonction remote_groupes pour %s'%groupname)
        return ldap_utils.get_groupes(groupname)

    def xmlrpc_remote_membres(self, groupe, rgroupe=None):
        """renvoie la liste des membres de 'groupe'
        """
        log.msg('Appel de la fonction remote_membres pour %s (filtre : %s)'%(groupe, rgroupe))
        if rgroupe: return ldap_utils.get_membres_croise([groupe, rgroupe])
        else: return ldap_utils.get_membres(groupe)

    def xmlrpc_remote_get_devoirs(self, cle, username):
        """renvoie la liste des devoirs de 'username'
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_get_devoirs pour %s'%(username))
        return devoirs.get_devs(username)

    def xmlrpc_remote_distribute(self, cle, username, groupe, dev_name, eleve_only, in_perso):
        """distribue "dev_name" du prof "username" à "groupe"
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_distribute pour : username=%s, groupe=%s, dev_name=%s, eleve_only=%s, in_perso=%s'%(username, groupe, dev_name, eleve_only, in_perso))
        return devoirs.distribute(username, groupe, dev_name, eleve_only, in_perso)

    def xmlrpc_remote_ramasse(self, cle, username, dev_name):
        """rammasse le devoir "dev_name" du prof "username"
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_ramasse pour : username=%s, dev_name=%s'%(username, dev_name))
        return devoirs.ramasse(username, dev_name)

    def xmlrpc_remote_rendre(self, cle, username, dev_name):
        """rend la correction du devoir "dev_name" du prof "username"
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_rendre pour : username=%s, dev_name=%s'%(username, dev_name))
        return devoirs.rendre(username, dev_name)

    def xmlrpc_remote_supprimer(self, cle, username, dev_name):
        """Supprime les données du devoir "dev_name" du prof "username"
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_supprimer pour : username=%s, dev_name=%s'%(username, dev_name))
        return devoirs.supprimer(username, dev_name)

    def xmlrpc_remote_copie_fichier(self, cle, username, dev_name, emplacement, nom_fichier, donnees_fichier):
        """Réception et écriture d'un fichier vers /home/workgroups/devoirs/<username>/<dev_name> selon son <emplacement> ("devoir" ou "data", "devoir" = considéré comme valeur par défaut)
           Si <emplacement> = "devoir" alors écriture dans /home/workgroups/devoirs/<username>/<dev_name>
           Si <emplacement> = "data" alors écriture dans /home/workgroups/devoirs/<username>/<dev_name>/donnees
           <donnees_fichier> = objet xmlrpclib.Binary
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return False
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return False

        log.msg('Appel de la fonction remote_copie_fichiers_devoir pour : username=%s, dev_name=%s, emplacement=%s, nom_fichier=%s'%(username, dev_name, emplacement, nom_fichier))

        # emplacement = "devoir" est considéré comme valeur par défaut
        dirprofdev = os.path.join("/home/workgroups/devoirs", username, dev_name)
        if not os.path.isdir(dirprofdev): os.makedirs(dirprofdev)

        if emplacement == "data":
            dirprofdev = os.path.join(dirprofdev, "donnees")
            if not os.path.isdir(dirprofdev): os.makedirs(dirprofdev)

        chemin_fichier = os.path.join(dirprofdev, nom_fichier)

        with open(chemin_fichier, "wb") as handle:
            handle.write(donnees_fichier.data)

        os.system('setfacl -PRm u:%s:rwx %s'%(username, dirprofdev))

        return True

    def xmlrpc_remote_get_dev_todel(self, cle, username):
        """Retourne la liste des devoirs pour un prof dont les données sont tjrs sur le serveur
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction remote_get_dev_todel pour %s'%(username))

        dirprofdev = os.path.join("/home/workgroups/devoirs", username)
        l = []
        if not os.path.exists(dirprofdev): return l
        for dev_name in os.listdir(dirprofdev):
            if dev_name.lower() == 'tmp': continue
            if os.path.isdir(os.path.join(dirprofdev, dev_name)):
                l.append('%s'%(dev_name))
        l.sort()
        return l

    def xmlrpc_remote_get_mail(self, cle, username):
        """Renvoi le mail de cet utilisateur
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []

        log.msg('Appel de la fonction xmlrpc_remote_get_mail pour %s'%(username))

        conn = Ldap()
        conn.connect()
        ldapuser = _user_factory(username, conn.connexion)
        return ldapuser.get_attr(username, "mail")[0]

    def xmlrpc_remote_get_nom(self, cle, username):
        """Renvoi le titre et le nom de cet utilisateur (ex: M. Trucmuche)
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return []

        log.msg('Appel de la fonction xmlrpc_remote_get_nom pour %s'%(username))

        conn = Ldap()
        conn.connect()
        ldapuser = _user_factory(username, conn.connexion)
        return "%s %s" % (ldapuser.get_attr(username, "personalTitle")[0], ldapuser.get_attr(username, "sn")[0])

    def xmlrpc_remote_check_nb_min_classes(self, cle, password, nb_classe_min):
        """Teste si le password contient bien nb_classe_min de classe de caractères
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return []

        log.msg('Appel de la fonction xmlrpc_remote_check_nb_min_classes')

        return check_nb_min_classes(password, nb_classe_min)

    def xmlrpc_remote_change_mdp(self, cle, username, old_password, new_password):
        """Change le mot de passe de username en vérifiant son ancien mdp
        """
        if not self.verif_cle(cle) :
            log.msg('Clé erronnée : %s'%cle)
            return False
        elif not self.est_prof(username) :
            log.msg('User non prof : %s'%username)
            return False

        log.msg('Appel de la fonction xmlrpc_remote_change_mdp')

        conn = Ldap()
        conn.connect()
        ldapuser = _user_factory(username, conn.connexion)

        # si un prof change son mdp, il doit connaitre le précédent
        if not ldapuser.auth(username, old_password):
            return False

        # application du changement de mot de passe
        ldapuser.c_mod_password(username, new_password)
        conn.close()
        return True

#######################
# programme principal #
#######################
if __name__ == '__main__':
    sys.exit("Usage : twistd -noy %s"%os.path.basename(sys.argv[0]))

# Après un arrêt de controle-vnc, purge des stations allumees
manage_stations.purge()
application = service.Application('ControleVNCserveur')
# serveur de téléchargement de client Scribe
root = static.File('/home/client_scribe')
site = server.Site(root)
log.msg('Controle VNC Serveur - Fichiers demarre port:8790')
internet.TCPServer(8790, site).setServiceParent(service.IServiceCollection(application))
# serveur PB pour exécution de fonctions distantes
pbfactory = pb.PBServerFactory(Echoer(), unsafeTracebacks=True)
#pbfactory = pb.PBServerFactory(Echoer(), unsafeTracebacks=False)
log.msg('Controle VNC Serveur demarre port:8789')
internet.TCPServer(8789, pbfactory).setServiceParent(service.IServiceCollection(application))
#PBServerFactory.unsafeTracebacks

# serveur xmlrpc pour EOP (gestion de devoirs)
if activer_eop == 'oui':
    rpcserver = server.Site(RpcServer())
    log.msg('Controle VNC Serveur demarre port:8788')
    internet.TCPServer(8788, rpcserver).setServiceParent(service.IServiceCollection(application))
